-- hale_07_tracking.lua
-- hale telescope target tracking routines

-- the xyz to RA,Dec routines were taken from Lua Edu's celutil.lua
-- Thanks, Vincent!

--- all other code  Selden Ball, October, 2007. All rights reserved. 

-- useful constants:
local pkg = {
	LOOK  = celestia:newvector(0,0,-1);
	XAxis = celestia:newvector(1,0,0);
	YAxis = celestia:newvector(0,1,0);
	ZAxis = celestia:newvector(0,0,1);
	Second = 1.0/(24.0*60*60)
}

-- Transform coordinates from cartesian to polar 
pkg.transform_xyz2rtp = function (x,y,z)
  local r = math.sqrt(x*x + y*y + z*z)
  local theta = math.atan(math.sqrt(x*x + y*y)/z)
  local phi
  if x < 0 then
    phi = math.atan(y/x) + math.pi
  elseif x > 0 then
    phi = math.atan(y/x)
  elseif y > 0 then
    phi = math.pi/2
  else
    phi = -math.pi/2
  end
  return r,theta,phi
end


-- Return current Ra, Dec for selection
pkg.get_ra_dec = function (vp, sel)

  local base_rot = celestia:newrotation(celestia:newvector(1,0,0), -math.rad(23.4392911))
  local rot = vp:getposition():orientationto(sel:getposition(), pkg.LOOK) * base_rot

  local look = rot:transform(pkg.LOOK):normalize()
  local r,theta,phi = pkg.transform_xyz2rtp(look.x, look.z, look.y)
  local phi = math.mod(720 - math.deg(phi), 360)
  local theta = math.deg(theta)
  if theta > 0 then
    theta = 90 - theta
  else
    theta = (-90 - theta)
  end
  return phi, theta
end

--======================================
-- yoke's RA


prevRA  = nil
prevRAt = nil

getRAq = function (vp,date)

-- get desired operation (should be done in earliest periodic function)
hale_cmd(date)

-- avoid math errors if selection is (near) self

	local myPos = vp:getposition()

	local sel = celestia:getselection()
	local selPos = sel:getposition()
	local distance = myPos:distanceto(selPos)

	if distance < 0.01 then
--	if too close, point toward default target
		sel = celestia:find("Sol/Earth/hale_default_target")
	end

-- set initial pointing values immediately
	if ((prevRA == nil) or (date < prevRAt))
	then  
	  	local prevDecl
	  	prevRA,prevDecl  = pkg.get_ra_dec (vp,sel)
	  	prevRAt = date

		local rad_ra = math.rad(prevRA)
		local q_rotate1 = celestia:newrotation(pkg.YAxis,-rad_ra)	
		local q_rotate2 = celestia:newrotation(pkg.YAxis, math.rad(-90))
		local q_rotate = q_rotate1 * q_rotate2
		return q_rotate.w, q_rotate.x, q_rotate.y, q_rotate.z
		end

-- set bounds on math
	if     (prevRA < 0.0)   then prevRA = 360.0 + prevRA
	elseif (prevRA > 360.0) then prevRA = prevRA - 360.0 end

	local destinationRA0,dec = pkg.get_ra_dec (vp,sel)

	local v = 18.0/second
	local stepAngle = v * (date-prevRAt)

-- find closest dest: avoid rotating > 180 degrees
	if (destinationRA0 - prevRA > 180) 
	then destinationRA0 = destinationRA0 - 360
	elseif (prevRA - destinationRA0 > 180) 
	then destinationRA0 = destinationRA0 + 360 
	end


-- take baby steps toward target
	if (math.abs(destinationRA0 - prevRA) > stepAngle) then 
	  local thisStep = stepAngle
	  if (destinationRA0 < prevRA) then thisStep = - stepAngle end
	  destinationRA = prevRA +thisStep
	else destinationRA = destinationRA0
	end


-- 	s = "prevRA = "..prevRA..", stepAngle = "..stepAngle..", destinationRA0 = "..destinationRA0..", destinationRA = "..destinationRA
-- 	celestia:flash(s,1)


-- all was being done in degrees. Celestia wants radians
	local rad_ra = math.rad(destinationRA)

	local q_rotate1 = celestia:newrotation(pkg.YAxis,-rad_ra)	
	local q_rotate2 = celestia:newrotation(pkg.YAxis, math.rad(-90))
	local q_rotate = q_rotate1 * q_rotate2

-- keep track of where we are now
	prevRA = math.fmod(destinationRA,360)
	prevRAt = date

	return q_rotate.w, q_rotate.x, q_rotate.y, q_rotate.z

end

-- motion functions per Celestia's documentation

YokeRAproto =
 { 
	Viewpoint = "Sol/Earth/hale_yoke",
	Period = 1,
 }

-- constructor method

function YokeRAproto:new(o)
   o = o or {}  -- create table if one not provided
   setmetatable(o, self)
   self.__index = self

--   o.period = o.Period

   return o
end


-- The orientation function. 

function YokeRAproto:orientation(tjd)

	local myVP = celestia:find(self.Viewpoint)

   -- return a quaternion representing the current orientation
	return getRAq(myVP,tjd)
end

function YokeRA(sscvals)
   -- create a new rotation object
   return YokeRAproto:new(sscvals)
end

--================================================
-- telescope tube's declination angle

prevDecl  = nil
prevDeclT = nil

getDeclq = function (vp, date)

	local myPos = vp:getposition()

	local sel = celestia:getselection()
	local selPos = sel:getposition()
	local distance = myPos:distanceto(selPos)

--	if too close, point toward default target
	if distance < 0.01 then
	   sel = celestia:find("Sol/Earth/hale_default_target")		
	end

-- if no initial values yet, set to current target immediately
	if (prevDecl == nil) or (date < prevDeclT)
	  then
		local ra,dec = pkg.get_ra_dec (vp,sel)
	        prevDecl  = dec
		prevDeclT = date
		local rad_dest = math.rad(dec)
		local q_rotate2 = celestia:newrotation(pkg.YAxis, math.pi)
		local q_rotate1 = celestia:newrotation(pkg.XAxis,rad_dest)
		local q_rotate = q_rotate1  * q_rotate2
		return q_rotate.w, q_rotate.x, q_rotate.y, q_rotate.z
	end

-- determine dec of current target
	local ra,destination = pkg.get_ra_dec (vp,sel)

	local v = 18.0/second
	local stepAngle = v * (date-prevDeclT)

	if (math.abs(destination - prevDecl ) > stepAngle) then 
-- 	  only go partway: 
	  local thisStep = stepAngle
	  if (destination  < prevDecl ) then thisStep = - stepAngle end
	  destination  = prevDecl  +thisStep
	end

-- keep angle in range
	if (destination < -90.0)   
	  then destination = math.abs(destination+90.0) - 90.0
	elseif (destination > 90.0)
	  then destination = 90.0 - (destination-90.0) 
	end

--	s = "prev = "..prevDecl..", dest = "..destination.." = "..dec
--	celestia:flash(s,1)

	local rad_dest = math.rad(destination)

	local q_rotate2 = celestia:newrotation(pkg.YAxis, math.pi)
	local q_rotate1 = celestia:newrotation(pkg.XAxis,rad_dest)
	local q_rotate = q_rotate1  * q_rotate2

	prevDecl = destination
	prevDeclT = date

	return q_rotate.w, q_rotate.x, q_rotate.y, q_rotate.z
end

TelescopeDeclproto =
 { 
	Viewpoint = "Sol/Earth/Hale_position",
	Period = 1,
 }

-- constructor method

function TelescopeDeclproto:new(o)
   o = o or {}  -- create table if one not provided
   setmetatable(o, self)
   self.__index = self

--   o.period = o.Period

   return o
end


-- The orientation function. 

function TelescopeDeclproto:orientation(tjd)

	local myVP = celestia:find(self.Viewpoint)

   -- return a quaternion representing the orientation
	return getDeclq(myVP,tjd)
end

function TelescopeDecl(sscvals)
   -- create a new rotation object
	return TelescopeDeclproto:new(sscvals)
end
